/*
 *	The Lestrade Search Library -- Operations with Indices
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 *
 */

#include <lestrade/lib.h>
#include <ucw/io.h>
#include <lestrade/internal.h>
#include <lestrade/index.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

int
lestrade_open_index(byte *filename, struct lestrade_index **pidx)
{
  struct lestrade_index *idx = xmalloc_zero(sizeof(*idx));
  strncpy(idx->name, filename, sizeof(idx->name)-1);
  int err;

  ucw_stat_t st;
  int fd = open(filename, O_RDONLY);
  if (fd < 0)
    {
      err = LESTRADE_ERR_INDEX_NOT_FOUND;
      msg(L_ERROR, "Cannot open: %m");
      goto bad;
    }
  if (ucw_fstat(fd, &st) < 0)
    {
      err = LESTRADE_ERR_INDEX_STAT_FAILED;
      msg(L_ERROR, "Cannot stat: %m");
      goto bad1;
    }
  if (st.st_size < (int)sizeof(struct index_header))
    {
      err = LESTRADE_ERR_INDEX_MALFORMED;
      goto bad1;
    }
  idx->map_length = st.st_size;
  idx->map_start = mmap(NULL, idx->map_length, PROT_READ, MAP_SHARED, fd, 0);
  if (idx->map_start == MAP_FAILED)
    {
      err = LESTRADE_ERR_INDEX_MMAP_FAILED;
      goto bad1;
    }

  idx->hdr = (struct index_header *) idx->map_start;
  struct index_header *hdr = idx->hdr;
  if (hdr->magic == INDEX_TEMP_MAGIC)
    {
      err = LESTRADE_ERR_INDEX_NOT_FINISHED;
      goto bad2;
    }
  if (hdr->magic != INDEX_MAGIC)
    {
      err = LESTRADE_ERR_INDEX_MALFORMED;
      goto bad2;
    }
  if (hdr->version != INDEX_VERSION)
    {
      err = LESTRADE_ERR_INDEX_BAD_VERSION;
      goto bad2;
    }
  idx->cih = (struct custom_index_hdr *)(idx->map_start + hdr->custom_start);
  idx->attrs = (struct card_attrs *)(idx->map_start + hdr->card_attr_start);
  idx->refs = idx->map_start + hdr->refs_start;
  idx->lexicon = idx->map_start + hdr->lexicon_start;
  idx->lexicon_ptrs = (u32*)(idx->map_start + hdr->lexicon_ptrs_start);
  idx->lexicon_zones = hdr->lexicon_zone;

  close(fd);
  *pidx = idx;
  return 0;

 bad2:
  munmap(idx->map_start, idx->map_length);
 bad1:
  close(fd);
 bad:
  xfree(idx);
  return err;
}

void
lestrade_close_index(struct lestrade_index *idx)
{
  munmap(idx->map_start, idx->map_length);
  xfree(idx);
}

struct custom_index_hdr *
lestrade_get_custom_index_hdr(struct lestrade_index *idx)
{
  return idx->cih;
}
