/*
 *	The Lestrade Search Library -- Index Generator
 *
 *	(c) 2004--2005 Martin Mares <mj@ucw.cz>
 *	(c) 2004--2005 Tomas Valla <tom@ucw.cz>
 *	(c) 2009 Pavel Charvat <pchar@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#undef LOCAL_DEBUG

#include <lestrade/lib.h>
#include <lestrade/internal.h>
#include <lestrade/index.h>
#include <ucw/fastbuf.h>
#include <ucw/ff-unicode.h>
#include <ucw/ff-binary.h>
#include <charset/unicat.h>
#include <ucw/unicode.h>
#include <ucw/conf.h>
#include <time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <lestrade/tagged-text.h>

/* The main buffer for storing (and sorting) gathered references */
struct refrec {
	u32 wordid;
	u32 pos;	// in the highest 4 bits we remember word-types
	u32 oid;
};
static struct refrec *ref_buf;

/* Auxiliary buffer, used when ref_buf is full */
static struct refrec *ref_buf_extra;

static uint ref_buf_cnt;
static uint ref_buf_size = 3*1024*1024;
static uint ref_buf_extra_size = 128*1024;

/* Buffer for storing parts of reference chains */
static byte *ref_chain;
static uint ref_chain_size = 65536;

/* When ref_buf is full, temporary file is created here and ref_buf dumped (in indexfile format) */
static struct fastbuf *ref_fb;

/*
 * Words in the lexicon (indexed by both original ID's [1..lex_cnt] and new ID's [0..lex_cnt-1])
 * In each stage of indexing, (struct verbum).(union u) contains data specific for that stage:
 *	len		Length of the word in characters [in make_lex_buf()]
 *	wid		New word ID [set by make_lex_buf() and used for sorting reference chains]
 *	refch_start	Start of the reference chain for this word [set when writing refchains]
 */
static struct verbum **lex_old, **lex_new;
static uint lex_cnt;

/* Current indexfile output */
static struct fastbuf *idx;

static uint curr_oid;

static struct index_header index_hdr;

#define LH_MKLEX
#define LH_NEED_CLEANUP
#include <lestrade/lexhash.h>

static int indexer_verbose;
#define TRACE(x...) do { if (indexer_verbose) msg(L_DEBUG, x); } while(0)

static struct cf_section indexer_config = {
	CF_ITEMS {
		CF_INT("Verbose",		&indexer_verbose),
		CF_UNS("RefBufSize",		&ref_buf_size),
		CF_UNS("RefBufExtraSize",	&ref_buf_extra_size),
		CF_UNS("RefChainSize",		&ref_chain_size),
		CF_END
	}
};

static void CONSTRUCTOR indexer_cfg_init(void)
{
	cf_declare_section("Indexer", &indexer_config, 0);
}

static void write_tmp_header(void)
{
	struct index_header hdr;

	DBG("Writing temporary header...");
	bzero(&hdr, sizeof(hdr));
	hdr.magic = INDEX_TEMP_MAGIC;
	hdr.version = INDEX_VERSION;
	hdr.timestamp = time(NULL);

	bwrite(idx, &hdr, sizeof(hdr));
	DBG("Done.");
}

static enum word_class lm_lookup(enum word_class orig_class, u16 *uni, uint ulen, word_id_t *idp)
{
	struct verbum *v;

	DBG("lm_lookup");
	if (orig_class != WC_NORMAL)
		return orig_class;
	v = lh_lookup(uni, ulen);
	*idp = v;
	return v->id & 7;
}

static inline int refrec_lt(struct refrec a, struct refrec b)
{
	COMPARE_LT(a.wordid, b.wordid);
	COMPARE_LT(a.oid, b.oid);
	COMPARE_LT(a.pos & 0xfffffff, b.pos & 0xfffffff);
	return 0;
}

#define ASORT_PREFIX(x) ref_##x
#define ASORT_KEY_TYPE struct refrec
#define ASORT_ELT(i) buf[i]
#define ASORT_LT(x,y) refrec_lt(x,y)
#define ASORT_EXTRA_ARGS , struct refrec *buf
#include <ucw/sorter/array-simple.h>

/*
 * We have to define another sorter which is used when ref_buf was never flushed into tempfile.
 * We need to compare according to the new word numbering.
 */
static inline int refrec2_lt(struct refrec a, struct refrec b)
{
	COMPARE_LT(lex_old[a.wordid]->u.wid, lex_old[b.wordid]->u.wid);
	COMPARE_LT(a.oid, b.oid);
	COMPARE_LT(a.pos & 0xfffffff, b.pos & 0xfffffff);
	return 0;
}

#define ASORT_PREFIX(x) ref2_##x
#define ASORT_KEY_TYPE struct refrec
#define ASORT_ELT(i) buf[i]
#define ASORT_LT(x,y) refrec2_lt(x,y)
#define ASORT_EXTRA_ARGS , struct refrec *buf
#include <ucw/sorter/array-simple.h>

/*
 * Encode data to indexfile format and write them to fb.
 * If fb==idx, wordid is omitted and refchain starts are computed.
 */
static void ref_buf_dump(struct fastbuf *fb, int phase2)
{
	u32 hdr, loid, lwordid, wid, wt, pos;
	uint size;
	uint i, rp, rep;

	DBG("Dumping ref_buf");

	loid = lwordid = 0;
	hdr = size = 0;
	rp = rep = 0;

	while ((i = rp + rep) < ref_buf_cnt) {

		/* We perform the ref_buf and ref_buf_extra merging here */
		struct refrec rb;
		if (rep + ref_buf_size >= ref_buf_cnt || (rp < ref_buf_size &&
				(phase2 ? refrec2_lt : refrec_lt)(ref_buf[rp], ref_buf_extra[rep])))
			rb = ref_buf[rp++];
		else
			rb = ref_buf_extra[rep++];

		wid = rb.wordid;
		DBG("wordid=%x, oid=%x, pos=%x", wid, rb.oid, rb.pos);
		if (wid != lwordid) {
			DBG("New word %d <%s>", wid, phase2 ? (const char *)lex_old[wid]->word : "?");
			lwordid = wid;
			ASSERT(wid);

			if (i>0) {
				if (size <= 15)
					hdr |= size;
				bputl(fb, hdr);
				if (size > 15)
					bput_utf8_32(fb, size);
				bwrite(fb, ref_chain, size);
				if (phase2)
					bputl(fb, 0);
			}

			if (phase2)
				lex_old[wid]->u.refch_start = btell(idx) - index_hdr.refs_start;

			loid = rb.oid;
			size = 0;
			hdr = rb.oid << 4;
			ASSERT(hdr);

			/* We need wordid (only) in tmpfile */
			if (!phase2)
			  bputl(fb, wid);
		}
		else if (rb.oid != loid) {
			DBG("New object %d", rb.oid);
			loid = rb.oid;
			if (i>0) {
				if (size <= 15)
					hdr |= size;
				bputl(fb, hdr);
				if (size > 15)
					bput_utf8_32(fb, size);
				bwrite(fb, ref_chain, size);
			}

			size = 0;
			hdr = rb.oid << 4;
			ASSERT(hdr);

			if (!phase2)
			  bputl(fb, wid);
		}

		if (unlikely(size+3 >= ref_chain_size)) {
			DBG("ref_chain realloc");
			ref_chain_size *= 2;
			ref_chain = xrealloc(ref_chain, ref_chain_size);
		}

		/*
		 * Reference encoding follows
		 */
		wt = rb.pos >> 28;
		pos = rb.pos & 0xfffffff;
		if (wt <= 1 && pos <= 0x3f)
			ref_chain[size++] = (wt << 6) | pos;
		else
		if (wt <= 7 && pos <= 0x7ff) {
			ref_chain[size++] = 0x80 | (wt << 3) | (pos >> 8);
			ref_chain[size++] = pos & 0xff;
		} else {
			ref_chain[size++] = 0xc0 | (wt << 2) | (pos >> 16);
			ref_chain[size++] = (pos >> 8) & 0xff;
			ref_chain[size++] = pos & 0xff;
		}
		DBG("reference size=%d", size);
	}

	/* And dump the rest ... */
	if (i>0) {
		if (size <= 15)
			hdr |= size;
		bputl(fb, hdr);
		if (size > 15) {
			bput_utf8_32(fb, size);
		}
		bwrite(fb, ref_chain, size);
		if (phase2)
			bputl(fb, 0);
	}
	DBG("Dumping finished.");
}

static struct fastbuf empty_fb;		// Only returns EOF
static struct fastbuf *scan_char_fb;
#define LM_SCAN_CHAR(u) ( ((u)=bget_tagged_char(scan_char_fb)) == ~0U ? 0 : 1)
// #define LM_SCAN_CHAR(u) ( ((u)=bget_utf8(scan_char_fb)) == ~0U ? 0 : 1)

static void insert_ref(u32 wid, u32 pos, u32 oid, uint cat)
{
	if (pos > 0x3ffff) {
		msg(L_WARN, "Too big document encountered (oid=%x, wordid=%x, pos=%x), skipping the rest.", oid, wid, pos);
		scan_char_fb = &empty_fb;
		return;
	}

	if (unlikely(ref_buf_cnt >= ref_buf_size+ref_buf_extra_size)) {
		DBG("ref_buf_extra realloc");
		ref_buf_extra_size *= 2;
		ref_buf_extra = xrealloc(ref_buf_extra, ref_buf_extra_size * sizeof(*ref_buf_extra));
	}

	if (ref_buf_cnt >= ref_buf_size)
		ref_buf_extra[ref_buf_cnt-ref_buf_size] = (struct refrec){ .wordid = wid, .pos = pos | (cat<<28), .oid = oid };
	else
		ref_buf[ref_buf_cnt] = (struct refrec){ .wordid = wid, .pos = pos | (cat<<28), .oid = oid };

	ref_buf_cnt++;
}

static void lm_got_word(uint pos, uint cat, word_id_t w)
{
	DBG("Got word, pos=%d cat=%d id=%d class=%d", pos, cat, w->id >> 3, w->id & 3);

	insert_ref(w->id >> 3, pos, curr_oid, cat);
}

static void lm_got_complex(uint pos UNUSED, uint cat UNUSED, word_id_t root UNUSED, word_id_t w UNUSED, uint dir UNUSED)
{
	DBG("Got complex");
	ASSERT(0);
}

#include <lestrade/lexmap.h>


static void chew_data(struct fastbuf *fb)
{
	scan_char_fb = fb;

	DBG("Chewing data of oid=%x", curr_oid);
	lm_map_text();
	if (ref_buf_cnt >= ref_buf_size) {
		DBG("Ref_buf is full (%d items, %d in ref_buf_extra)", ref_buf_cnt, ref_buf_cnt-ref_buf_size);
		if (!ref_fb) {
			TRACE("Out of buffers, writing references to a temporary file");
			ref_fb = bopen_tmp(65536);
		}

		DBG("Sorting ref_buf (%d items)", ref_buf_cnt);
		if (ref_buf_cnt > ref_buf_size) {
			ref_sort(ref_buf_size, ref_buf);
			ref_sort(ref_buf_cnt - ref_buf_size, ref_buf_extra);
		}
		else
			ref_sort(ref_buf_cnt, ref_buf);

		DBG("Finished sorting");
		ref_buf_dump(ref_fb, 0);
		ref_buf_cnt = 0;
	}

}

static inline int lexword_lt(struct verbum *x, struct verbum *y)
{
	/* Phase 1: lengths */
	COMPARE_LT(x->u.len, y->u.len);

	/* Phase 2: words without accents */
	const byte *xx = x->word, *yy = y->word;
	uint u1, u2;
	while (*xx && *yy) {
		xx = utf8_get(xx, &u1);
		yy = utf8_get(yy, &u2);
		u1 = Uunaccent(u1);
		u2 = Uunaccent(u2);
		COMPARE_LT(u1, u2);
	}

	/* Phase 3: words with accents */
	if (strcmp(x->word, y->word)<0)
		return 1;
	return 0;
}

#define ASORT_PREFIX(x) lex_##x
#define ASORT_KEY_TYPE struct verbum *
#define ASORT_ELT(i) lex_new[i]
#define ASORT_LT(x,y) lexword_lt(x,y)
#include <ucw/sorter/array-simple.h>

static void write_lexicon(void)
{
	uint l,i;
	u32 lp;
	uint zone=0;
	
	DBG("Writing lexicon pointers into index...");
	index_hdr.lexicon_ptrs_start = btell(idx);
	bputl(idx, 0);
	for (i=0, lp=0; i<lex_cnt; i++) {
		l = utf8_strlen(lex_new[i]->word);
		ASSERT(l);
		while (zone <= l)
			index_hdr.lexicon_zone[zone++] = i;
		lp += 1 + strlen(lex_new[i]->word) + BYTES_PER_O;
		bputl(idx, lp);
	}
	while (zone <= MAX_WORD_CHARS+1)
		index_hdr.lexicon_zone[zone++] = i;
	DBG("Done.");
		
	DBG("Writing lexicon into index...");
	index_hdr.lexicon_start = btell(idx);
	for (i=0; i<lex_cnt; i++) {
		struct verbum *v = lex_new[i];
		DBG("<%s>", v->word);
		bputc(idx, strlen(v->word));
		bputs(idx, v->word);
		bputo(idx, v->u.refch_start);
	}
	DBG("Done.");
}

struct reffb_key {
  u32 size;
  u32 wid;
  u32 hdr;
};

#define SORT_PREFIX(x) reffb_##x
#define SORT_KEY struct reffb_key
#define SORT_DATA_SIZE(k) ((k).size)
#define SORT_INPUT_FB
#define SORT_OUTPUT_FB

static inline int reffb_compare(struct reffb_key *a, struct reffb_key *b)
{
	COMPARE(lex_old[a->wid]->u.wid, lex_old[b->wid]->u.wid);
	COMPARE(a->hdr >> 4, b->hdr >> 4);
	return 0;
}

static inline int reffb_read_key(struct fastbuf *f, struct reffb_key *k)
{
	uint x = bgetl(f);
	if (!~x)
		return 0;
	k->wid = x;
	k->hdr = bgetl(f);
	if (k->hdr & 15)
		k->size = k->hdr & 15;
	else
		k->size = bget_utf8_32(f);
	return 1;
}

static inline void reffb_write_key(struct fastbuf *f, struct reffb_key *k)
{
	bputl(f, k->wid);
	bputl(f, k->hdr);
	if (!(k->hdr & 15))
		bput_utf8_32(f, k->size);
}

#include <ucw/sorter/sorter.h>

/*
 * Copy reference chains from tempfile to index and omit the wordid,
 * compute the refchain starts of each word to store them in lexicon
 */
static void write_refchains(struct fastbuf *f)
{
	u32 wid, hdr, owid = ~0U;

	DBG("Rewriting reference chains...");
	brewind(f);

	for (uint i=1; i<=lex_cnt; i++)
		lex_old[i]->u.refch_start = ~(index_pos_t)0;

	while (breadb(f, &wid, 4)) {
		if (wid != owid) {
			DBG("wid=%x", wid);
			if (owid != (u32)~0U)
				bputl(idx, 0);
			lex_old[wid]->u.refch_start = btell(idx) - index_hdr.refs_start;
			owid = wid;
		}
		hdr = bgetl(f);
		bputl(idx, hdr);
		uint size = hdr & 15;
		if (!size) {
			size = bget_utf8_32(f);
			bput_utf8_32(idx, size);
		}
		DBG("oid=%x size=%x", hdr >> 4, size);
		bbcopy(f, idx, size);
	}

	if (owid != (u32)~0U)
		bputl(idx, 0);
	DBG("Done.");
}

/* Create lexicon in the memory and sort it (see lex_lt()) */
static void make_lex_buf(void)
{
	lex_cnt = lh_id >> 3;
	lex_old = xmalloc((lex_cnt+1) * sizeof(struct verbum *));
	LH_WALK(v) {
		uint wid = v->id >> 3;
		DBG("%d: <%s>", wid, v->word);
		v->u.len = utf8_strlen(v->word);	// precompute word length
		ASSERT(wid && wid <= lex_cnt);
		lex_old[wid] = v;
	}
	DBG("Sorting lexicon wcnt=%d ...", lex_cnt);
	lex_new = xmalloc(lex_cnt * sizeof(struct verbum *));
	memcpy(lex_new, lex_old+1, lex_cnt * sizeof(struct verbum *));
	lex_sort(lex_cnt);
	DBG("Done.");

	for (uint i=0; i<lex_cnt; i++)
		lex_new[i]->u.wid = i;	// renumber words for sorter
}

static void
indexer_init(void)
{
	static int indexer_initialized;
	if (!indexer_initialized) {
		lm_init();
		fbbuf_init_read(&empty_fb, NULL, 0, 0);
		indexer_initialized = 1;
	}
	if (!ref_buf) {
		ref_buf = xmalloc(ref_buf_size * sizeof(*ref_buf));
		ref_buf_extra = xmalloc(ref_buf_extra_size * sizeof(*ref_buf_extra));
		ref_chain = xmalloc(ref_chain_size * sizeof(*ref_chain));
	}
	ref_buf_cnt = 0;
}

static void
indexer_cleanup(void)
{
	if (ref_buf) {
		xfree(ref_buf); ref_buf = NULL;
		xfree(ref_buf_extra); ref_buf_extra = NULL;
		xfree(ref_chain); ref_chain = NULL;
	}
}

static uint make_index(struct fastbuf *out, card_enumerator_t source, void *enum_priv, struct custom_index_hdr *custom_hdr, uint custom_hdr_len)
{
	lh_init();
	indexer_init();

	bzero(&index_hdr, sizeof(index_hdr));
	ref_fb = NULL;

	/* Write temporary header and custom header */
	idx = out;
	write_tmp_header();
	index_hdr.custom_start = btell(idx);
	bwrite(idx, custom_hdr, custom_hdr_len);

	/* Scan cards and also dump card_attrs */
	TRACE("Scanning cards");
	index_hdr.card_attr_start = btell(idx);
	struct fastbuf *fb;
	struct card_attrs ca;
	uint card_cnt = 0;
	curr_oid = 1;
	while (fb = source(&ca, enum_priv)) {
		bwrite(idx, &ca, sizeof(ca));	// Just dump the card_attrs, we won't need them anymore
		lm_doc_start();
		chew_data(fb);
		bclose(fb);
		card_cnt++;
		curr_oid++;
	}
	TRACE("Processed %d cards", card_cnt);

	make_lex_buf();
	TRACE("Found %d words", lex_cnt);

	index_hdr.refs_start = btell(idx);
	if (ref_fb) {
		if (ref_buf_cnt) {
			DBG("Working with the rest of ref_buf (%d items)", ref_buf_cnt);
			DBG("Sorting ref_buf (%d items)", ref_buf_cnt);
			if (ref_buf_cnt > ref_buf_size) {
				ref_sort(ref_buf_size, ref_buf);
				ref_sort(ref_buf_cnt - ref_buf_size, ref_buf_extra);
			}
			else
				ref_sort(ref_buf_cnt, ref_buf);
			DBG("Finished sorting");
			ref_buf_dump(ref_fb, 0);
		}
		TRACE("Sorting temporary file");
		brewind(ref_fb);
		struct fastbuf *tmp = reffb_sort(ref_fb, NULL);
		TRACE("Writing references");
		write_refchains(tmp);
		bclose(tmp);
	}
	else {				// We don't have to use external tempfile
		if (ref_buf_cnt) {
			DBG("Sorting ref_buf (%d items)", ref_buf_cnt);
			if (ref_buf_cnt > ref_buf_size) {
				ref2_sort(ref_buf_size, ref_buf);
				ref2_sort(ref_buf_cnt - ref_buf_size, ref_buf_extra);
			}
			else
				ref2_sort(ref_buf_cnt, ref_buf);

			TRACE("Writing references");
			for (uint i=1; i<=lex_cnt; i++)
				lex_old[i]->u.refch_start = ~(index_pos_t)0;
			ref_buf_dump(idx, 1);
		}
	}

	TRACE("Writing lexicon");
	write_lexicon();

	/* Write final header */
	TRACE("Writing index header");
	brewind(idx);
	index_hdr.magic = INDEX_MAGIC;
	index_hdr.version = INDEX_VERSION;
	index_hdr.num_cards = card_cnt;
	index_hdr.timestamp = time(NULL);
	bwrite(idx, &index_hdr, sizeof(index_hdr));
	bclose(idx);

	/* Free memory */
	xfree(lex_old);
	xfree(lex_new);
	lh_cleanup();

	return card_cnt;
}

uint lestrade_make_index(byte *filename, card_enumerator_t source, void *enum_priv, struct custom_index_hdr *custom_hdr, uint custom_hdr_len)
{
	TRACE("Building index %s", filename);
	struct fastbuf *out = bopen(filename, O_WRONLY | O_CREAT | O_TRUNC, 65536);
	uint cnt = make_index(out, source, enum_priv, custom_hdr, custom_hdr_len);
	indexer_cleanup();
	TRACE("Index built");
	return cnt;
}
