/*
 *	The Lestrade Search Library -- Detailed Fulltext Matching
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#undef LOCAL_DEBUG

#include <lestrade/lib.h>
#include <ucw/mempool.h>
#include <ucw/bbuf.h>
#include <lestrade/internal.h>
#include <lestrade/tagged-text.h>
#include <lestrade/match.h>

#define GBUF_PREFIX(x) wb_##x
#define GBUF_TYPE struct lestrade_word_pos
#include <ucw/gbuf.h>

struct detail_status {
  struct lestrade_matches matches;
  bb_t card_text;
  wb_t card_words;
  uint card_len;
  uint word_cnt;
  struct lestrade_word_pos last_word_pos;
};

static void
details_init(struct query *q)
{
  if (!q->details)
    {
      struct detail_status *d = mp_alloc(q->pool, sizeof(struct detail_status));
      q->details = d;
      bb_init(&d->card_text);
      wb_init(&d->card_words);
    }
}

void
details_cleanup(struct query *q)
{
  struct detail_status *d;
  if (d = q->details)
    {
      bb_done(&d->card_text);
      wb_done(&d->card_words);
      q->details = NULL;
    }
}

static void
load_card_text(struct query *q, struct card_attrs *attrs)
{
  struct detail_status *d = q->details;
  struct fastbuf *fb = q->Q->fetcher(attrs, q->Q->fetcher_priv);
  uint len = 0;
  uint more;
  byte *buf;
  DBG("Reading card contents");
  while (more = bdirect_read_prepare(fb, &buf))
    {
      DBG("\t%d byte chunk", more);
      bb_grow(&d->card_text, len+more);
      memcpy(d->card_text.ptr+len, buf, more);
      len += more;
      bdirect_read_commit(fb, buf+more);
    }
  bclose(fb);
  DBG("\tRead %d bytes", len);
  bb_grow(&d->card_text, len+1);
  d->card_text.ptr[len] = 0;
  d->card_len = len;
}

static struct detail_status *fh_details;
typedef struct verbum *word_id_t;
#define LM_MAP_ARGS const byte *text
#define LM_SCAN_CHAR(u) (GET_TAGGED_CHAR(text,u), u)
#define LM_TRACK_TEXT

static enum word_class lm_lookup(enum word_class orig_class, u16 *uni, uint ulen, word_id_t *idp, const byte *ostart, uint olen)
{
  uint class = orig_class;

  if (class == WC_NORMAL)
    {
      uint seen_accents;
      struct verbum *v = fh_lookup_nonacc(uni, ulen, &seen_accents);
      if (v)
	{
	  if (!seen_accents)
	    *idp = (word_id_t)((uintptr_t)v | 1);	/* Yes, it's a hack, but I need to store that bit somewhere :) */
	  else
	    {
	      *idp = v;
	      if (v->has_accented_variants)
		{
		  struct verbum *va = fh_lookup(uni, ulen);
		  if (va)
		    {
		      *idp = (word_id_t)((uintptr_t)va | 1);
		      v = va;
		    }
		}
	    }
	  class = v->class;
	}
      else
	class = (ulen < lex_min_len_ign) ? WC_IGNORED : (ulen < lex_min_len) ? WC_GARBAGE : WC_NORMAL;
    }

  fh_details->last_word_pos = (struct lestrade_word_pos) {
    .pos = ostart - fh_details->card_text.ptr,
    .length = olen,
    .word_type = 0,
    .word_class = class,
    .word_index = 0xff
  };
  return class;
}

static inline void lm_got_word(uint pos, uint cat, word_id_t w)
{
  struct lestrade_word_pos *lp = &fh_details->last_word_pos;
  lp->word_type = cat;

  wb_grow(&fh_details->card_words, pos+1);
  while (fh_details->word_cnt < pos)
    {
      struct lestrade_word_pos *wp = &fh_details->card_words.ptr[fh_details->word_cnt];
      *wp = *lp;
      wp->length = 0;
      wp->word_class = 0xff;
      DBG(">> %d: <%.*s> t=%d c=%d w=%d", fh_details->word_cnt, wp->length, fh_details->card_text.ptr + wp->pos, wp->word_type, wp->word_class, wp->word_index);
      fh_details->word_cnt++;
    }
  struct lestrade_word_pos *wp = &fh_details->card_words.ptr[pos];
  *wp = *lp;

  if (w)
    {
      struct verbum *v = (word_id_t)((uintptr_t)w & ~(uintptr_t)1);
      uint precise = (uintptr_t)w & 1;
      DBG("@%d: <%s> prec=%d cat=%d", pos, v->word, precise, cat);
      for (uint i=0; i<v->num_occurs; i++)
	{
	  uint pen = (precise ? v->penalty : v->penalty_nonacc) [i];
	  if (pen < 255)
	    {
	      wp->word_index = v->word_idx[i];
	      trail_add(fh_query, pos, v->word_idx[i], cat, pen);
	    }
	}
    }
  DBG(">> %d: <%.*s> t=%d c=%d w=%d", pos, wp->length, fh_details->card_text.ptr + wp->pos, wp->word_type, wp->word_class, wp->word_index);
  fh_details->word_cnt++;
}

static inline void lm_got_complex(uint pos UNUSED, uint cat UNUSED, word_id_t root UNUSED, word_id_t w UNUSED, uint dir UNUSED)
{
  ASSERT(0);
}

#include <lestrade/lexmap.h>

int
lestrade_eval_detailed_init(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  query_trace(q, TRACE_GENERAL, "Detailed matcher: initializing");
  details_init(q);
  return full_init(q);
}

int
lestrade_eval_detailed_cleanup(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  int rc = full_cleanup(q);
  query_trace(q, TRACE_GENERAL, "Detailed matched: cleanup");
  return rc;
}

struct lestrade_matches *
lestrade_eval_detailed(struct lestrade_query *Q, struct card_attrs *attrs)
{
  struct query *q = Q->q;
  struct detail_status *d = q->details;
  struct lestrade_matches *m = &d->matches;
  ASSERT(d);

  DBG("--- SNIP ---");
  full_cnt++;
  load_card_text(q, attrs);
  lm_doc_start();
  trail_init(q, 1);
  matches_record_start(q);
  d->word_cnt = 0;
  fh_details = d;
  DBG("Lexmapping:");
  lm_map_text(d->card_text.ptr);
  wb_grow(&d->card_words, d->word_cnt+1);
  d->card_words.ptr[d->word_cnt] = (struct lestrade_word_pos) {
    .pos = d->card_len, .length = 0, .word_type = 0xff, .word_class = 0xff, .word_index = 0xff
  };
  m->quality = match_single(q, attrs);
  matches_record_stop(q);

  query_trace(q, TRACE_GENERAL, "Detailed matching returned %d word entries and %d matches", d->word_cnt, q->matches_items);
  m->text = d->card_text.ptr;
  m->text_length = d->card_len;
  m->words = d->card_words.ptr;
  m->num_words = d->word_cnt;
  m->matches = q->matches;
  m->num_matches = q->matches_items;
  return m;
}

void
details_global_init(void)
{
  lm_init();
}
