/*
 *	The Lestrade Search Library -- Matching with Indices
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#undef LOCAL_DEBUG

#include <lestrade/lib.h>
#include <ucw/unaligned.h>
#include <ucw/unicode.h>
#include <ucw/heap.h>
#include <ucw/wildmatch.h>
#include <charset/unicat.h>
#include <lestrade/internal.h>
#include <lestrade/index.h>
#include <lestrade/match.h>
#include <stdlib.h>
#include <string.h>

struct ref_chain {
  u32 next_oid;
  const byte *chain;
  byte word_id;
  byte penalty;
};

static void
index_add_chain(struct query *q, struct word *w, struct variant *v, byte *exact, byte *chain, uint extra_penalty)
{
  uint penalty = v->penalty + extra_penalty;

  for (uint i=0; i<q->ref_cnt; i++)
    {
      struct ref_chain *r = &q->ref_chains[i];
      if (r->chain == chain && r->word_id == w->id)
	{
	  if (r->penalty > penalty)
	    {
	      query_trace(q, TRACE_INDEX, "\t<%s> for <%s>, penalty %d (improving previous entry)", exact, v->w, penalty);
	      r->penalty = penalty;
	    }
	  else
	    query_trace(q, TRACE_INDEX, "\t<%s> for <%s>, penalty %d (keeping penalty at %d)", exact, v->w, penalty, r->penalty);
	  return;
	}
    }

  query_trace(q, TRACE_INDEX, "\t<%s> for <%s>, penalty %d", exact, v->w, penalty);
  if (q->ref_cnt >= q->ref_max)
    {
      q->ref_max *= 2;
      q->ref_chains = xrealloc(q->ref_chains, q->ref_max * sizeof(struct ref_chain));
    }
  struct ref_chain *r = &q->ref_chains[q->ref_cnt++];
  r->chain = chain;
  r->word_id = w->id;
  r->penalty = penalty;
}

static uint
word_unaccent_utf8(byte *w, byte *to)
{
  uint u;
  byte *stop = to + MAX_WORD_BYTES;
  byte *buf = to;

  while (*w)
    {
      w = utf8_get(w, &u);
      u = Uunaccent(u);
      buf = utf8_put(buf, u);
      if (buf >= stop)
	return 0;
    }
  *buf = 0;
  return buf - to;
}

static void
lex_extract(struct lestrade_index *idx, uint id, byte *buf)
{
  const byte *p = idx->lexicon + idx->lexicon_ptrs[id];
  uint len = *p++;
  ASSERT(len <= MAX_WORD_BYTES);
  memcpy(buf, p, len);
  buf[len] = 0;
}

static void
lex_extract_noacc(struct lestrade_index *idx, uint id, byte *buf)
{
  const byte *p = idx->lexicon + idx->lexicon_ptrs[id];
  uint len = *p++;
  ASSERT(len <= MAX_WORD_BYTES);
  const byte *stop = p + len;
  while (p < stop)
    {
      uint u;
      p = utf8_get(p, &u);
      u = Uunaccent(u);
      buf = utf8_put(buf, u);
    }
  *buf = 0;
}

static index_pos_t
lex_extract_pos(struct lestrade_index *idx, uint id)
{
  byte *p = idx->lexicon + idx->lexicon_ptrs[id];
  p += 1 + p[0];
  return GET_O(p);
}

static uint
lex_find_first(struct lestrade_index *idx, uint len, byte *key)		/* Find first entry >=key (ignoring accents) */
{
  int l, r, m, res;
  byte w[MAX_WORD_BYTES+1];

  l = idx->lexicon_zones[len];
  r = idx->lexicon_zones[len+1] - 1;
  DBG("lex_find_first: len=%d, l=%d, r=%d", len, l, r);
  while (l < r)
    {
      m = (l+r)/2;
      lex_extract_noacc(idx, m, w);
      res = strcmp(w, key);
      DBG("(%d,%d,%d) <%s> %c <%s>", l, r, m, w, (res ? (res < 0) ? '<' : '>' : '='), key);
      if (res < 0)
	l = m + 1;
      else
	r = m;
    }
  return l;
}

static void
index_match_word(struct query *q, struct lestrade_index *idx, struct word *w)
{
  struct variant *v;
  uint found = 0;
  CLIST_WALK(v, w->variants)
    {
      uint len = utf8_strlen(v->w);
      byte vu[MAX_WORD_BYTES+3], wa[MAX_WORD_BYTES+1], wu[MAX_WORD_BYTES+1];
      if (!word_unaccent_utf8(v->w, vu))
	continue;
      uint i = lex_find_first(idx, len, vu);
      while (i < idx->lexicon_zones[len+1])
	{
	  lex_extract_noacc(idx, i, wu);
	  if (strcmp(vu, wu))
	    break;
	  lex_extract(idx, i, wa);
	  DBG("Considering <%s> <%s>", wa, wu);
	  int exact = !strcmp(v->w, wa);
	  if (v->ignore_accents || exact)
	    {
	      index_pos_t pos = lex_extract_pos(idx, i);
	      if (pos != ~(index_pos_t)0)
		{
		  index_add_chain(q, w, v, wa, idx->refs + lex_extract_pos(idx, i), (exact ? 0 : accent_penalty));
		  found++;
		}
	    }
	  i++;
	}
    }
  w->not_matchable = !found;
}

static void
index_match_wild(struct query *q, struct lestrade_index *idx, struct word *w)
{
  uint total_matches = 0;
  uint old, first, last;
  struct variant *v = (struct variant *) clist_head(&w->variants);
  ASSERT(v);

  byte *px = w->unaccented_w;
  for (uint i=0; i<w->first_wild; i++)
    px += utf8_encoding_len(*px);

  for (uint ilen=w->first_wild; ilen<=MAX_WORD_CHARS; ilen++)
    {
      old = *px;
      *px = 0;
      first = lex_find_first(idx, ilen, w->unaccented_w);
      (*(px-1))++;			/* We know 0xff is not a valid character code */
      last = lex_find_first(idx, ilen, w->unaccented_w);
      (*(px-1))--;			/* We know 0xff is not a valid character code */
      *px = old;
      DBG("Zone for length %d: [%d,%d)", ilen, first, last);

      /* Scan the zone */
      for (uint i=first; i < last; i++)
	{
	  byte lex[MAX_WORD_BYTES+1], lexu[2*MAX_WORD_BYTES+1];
	  uint penalty = 0;
	  lex_extract(idx, i, lex);
	  if (v->ignore_accents)
	    {
	      lex_extract_noacc(idx, i, lexu);
	      if (!wp_match(w->wild_unacc, lexu))
		{
		  DBG("\t%d <%s> <%s> not at all", i, lex, lexu);
		  continue;
		}
	      if (wp_match(w->wild, lex))
		DBG("\t%d <%s> <%s> acc", i, lex, lexu);
	      else
		{
		  DBG("\t%d <%s> <%s> nonacc", i, lex, lexu);
		  penalty = accent_penalty;
		}
	    }
	  else
	    {
	      if (!wp_match(w->wild, lex))
		{
		  DBG("\t%d <%s> <%s> !strict", i, lex, lexu);
		  continue;
		}
	      DBG("\t%d <%s> strict", i, lex);
	    }
	  index_pos_t pos = lex_extract_pos(idx, i);
	  if (pos != ~(index_pos_t)0)
	    {
	      if (total_matches >= wild_max_variants)
		query_error(q, LESTRADE_ERR_WILD_TOO_RICH, "Wildcard pattern matches too many words");
	      total_matches++;
	      index_add_chain(q, w, v, lex, idx->refs + lex_extract_pos(idx, i), penalty);
	    }
	}
    }
}

static void
index_find_words(struct query *q, struct lestrade_index *idx)
{
  for (uint i=0; i<q->num_words; i++)
    {
      struct word *w = q->words[i];
      if (w->word_class != WC_NORMAL)
	{
	  w->not_matchable = 1;
	  continue;
	}
      query_trace(q, TRACE_INDEX, "Searching for word <%s> and its variants in the lexicon", w->w);
      if (w->first_wild)
	index_match_wild(q, idx, w);
      else
	index_match_word(q, idx, w);
    }
}

static inline int
refs_lt(struct ref_chain *x, struct ref_chain *y)
{
  COMPARE_LT(x->chain, y->chain);
  COMPARE_LT(x->word_id, y->word_id);
  COMPARE_LT(x->penalty, y->penalty);
  return 0;
}

#define ASORT_PREFIX(x) refs_##x
#define ASORT_KEY_TYPE struct ref_chain
#define ASORT_LT(x,y) refs_lt(&(x), &(y))
#define ASORT_ELT(i) chains[i]
#define ASORT_EXTRA_ARGS ,struct ref_chain *chains
#include <ucw/sorter/array-simple.h>

#define RHEAP_LESS(a,b) (a->next_oid < b->next_oid)
#define RHEAP_SWAP(heap,a,b,t) (t=heap[a], heap[a]=heap[b], heap[b]=t)

static void
index_setup_chains(struct query *q)
{
  DBG("Optimizing refchains");
  if (q->ref_heap_size < q->ref_cnt+1)
    {
      q->ref_heap_size = q->ref_cnt + 1;
      q->ref_heap = xrealloc(q->ref_heap, q->ref_heap_size * sizeof(struct ref_chain *));
    }
  refs_sort(q->ref_cnt, q->ref_chains);
  struct ref_chain *r = q->ref_chains;
  struct ref_chain *end = r + q->ref_cnt;
  struct ref_chain **heap = q->ref_heap;
  q->ref_cnt = 0;
  while (r < end)
    {
      if (!q->ref_cnt || r->chain != heap[q->ref_cnt]->chain || r->word_id != heap[q->ref_cnt]->word_id)
	{
	  u32 x = GET_U32(r->chain);
	  if (x)
	    {
	      r->next_oid = x >> 4;
	      DBG("\tChain (w=%d,p=%d) starts with oid %08x", r->word_id, r->penalty, r->next_oid);
	      heap[++q->ref_cnt] = r;
	    }
	}
      r++;
    }
  HEAP_INIT(struct ref_chain *, heap, q->ref_cnt, RHEAP_LESS, RHEAP_SWAP);
}

static void
index_process_chains(struct query *q, struct lestrade_index *idx)
{
  struct ref_chain **heap = q->ref_heap;
  uint heap_cnt = q->ref_cnt;
  uint oid;

  while (heap_cnt)
    {
      oid = heap[1]->next_oid;
      DBG("Next OID is %x", oid);
      trail_init(q, 0);
      while (heap_cnt && heap[1]->next_oid == oid)
	{
	  struct ref_chain *r = heap[1];
	  const byte *ch = r->chain;
	  uint len = GET_U32(ch) & 0xf;
	  ch += 4;
	  if (!len)
	    ch = utf8_32_get(ch, &len);
	  DBG("\tChain for word %d, length %d", r->word_id, len);
	  const byte *stop = ch + len;
	  while (ch < stop)
	    {
	      uint pos = *ch;
	      uint wt;
	      if (pos < 0x80)
		{
		  wt = pos >> 6;
		  pos &= 0x3f;
		  ch++;
		}
	      else if (pos < 0xc0)
		{
		  wt = (pos >> 3) & 7;
		  pos = ((pos & 7) << 8) | ch[1];
		  ch += 2;
		}
	      else
		{
		  wt = (pos >> 2) & 15;
		  pos = ((pos & 3) << 16) | (ch[1] << 8) | ch[2];
		  ch += 3;
		}
	      trail_add(q, pos, r->word_id, wt, r->penalty);
	    }
	  ASSERT(ch == stop);
	  r->chain = ch;
	  u32 x = GET_U32(ch);
	  if (x)
	    {
	      r->next_oid = x >> 4;
	      DBG("\t\tnext OID %x", r->next_oid);
	      HEAP_INCREASE(struct ref_chain *, heap, heap_cnt, RHEAP_LESS, RHEAP_SWAP, 1, r);
	    }
	  else
	    {
	      DBG("\t\tEnd of chain.");
	      HEAP_DELETE_MIN(struct ref_chain *, heap, heap_cnt, RHEAP_LESS, RHEAP_SWAP);
	    }
	}
      match_single(q, &idx->attrs[oid-1]);
    }
}

static void
index_init(struct query *q)
{
  q->ref_cnt = 0;
  if (!q->ref_chains)
    {
      q->ref_max = 32;
      q->ref_chains = xmalloc(q->ref_max * sizeof(struct ref_chain));
      q->ref_heap = NULL;
    }
}

void
index_cleanup(struct query *q)
{
  xfree(q->ref_chains);
  q->ref_chains = NULL;
  xfree(q->ref_heap);
  q->ref_heap = NULL;
}

int
lestrade_eval_index(struct lestrade_query *Q, struct lestrade_index *idx)
{
  struct query *q = Q->q;
  query_trace(q, TRACE_GENERAL, "Index matcher: using index %s", idx->name);

  if (!setjmp(q->err_jmp))
    {
      index_init(q);
      index_find_words(q, idx);
      struct lestrade_expr *x = query_simplify(q, q->prep_expr);
      if (x->type == EX_NONE)
	query_trace(q, TRACE_GENERAL, "Shortcut: No matches possible");
      else
	{
	  index_setup_chains(q);
	  bool_prepare(q, x);
	  index_process_chains(q, idx);
	  bool_cleanup(q);
	}
    }
  query_trace(q, TRACE_GENERAL, "Index matcher: finished index %s", idx->name);
  return Q->error_code;
}
